/*****************************************************************************
    access.c

    Diese Datei ist Teil des Programmpakets 'ACL-tftp'.
    Alle Rechte bei Andreas G. Schindler
                    Klappacher Str. 2i
                    64285 Darmstadt, Deutschland.

    Dieses Programm ist 'Freie Software'; Sie knnen es entsprechend den 
    Bedingungen der 'Allgemeinen ffentlichen GNU-Lizenz' (verffentlicht 
    durch die 'Free Software Foundation') weiter verbreiten und/oder 
    modifizieren. Massgeblich ist die Version 2 oder eine neuere Version
    dieser Lizenz. 

    Dieses Programm wird in der Hoffnung dass es ntzlich is weitergegeben,
    jedoch OHNE JEDE GARANTIE. Dieser Garantieauschluss bezieht sich auch
    auf Eigenschaften wie Marktfhigkeit oder Eignung zu einem bestimmten
    Zweck. Details finden sich in der 'Allgemeinen ffentlichen GNU-Lizenz'.

    Eine Kopie dieser Lizenz sollte in  Programmpaket, dem diese Datei
    entstammt, enthalten sein. Falls nicht, schreiben Sie an die 
    'Free Software Foundation, Inc.', 59 Temple Place, Suite 330, Boston, 
    MA  02111-1307  USA.


    This file is part of the 'ACL-tftp' Package
    Copyright (C) 1999-2003 by Andreas G. Schindler
                               Klappacher Str. 2i
                               Darmstadt, Germany.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*****************************************************************************/
#include "auto.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include "params.h"
#include "globals.h"
#include "patstr.h"
#include "netio.h"
#define  PERMTABLES
#include "acl.h"
#include "dll.h"
#include "dllapps.h"
#include "util.h"
#include "logfun.h"
#include "readconfig.h"
#include "process.h"
#include "dump.h"
#include "access.h"

#define PERM_CA		(PERM_CREATE|PERM_APPEND)
#define PERM_RCA	(PERM_CREATE|PERM_APPEND|PERM_READ)
#define PERM_WCA	(PERM_CREATE|PERM_APPEND|PERM_WRITE)
#define PERM_RWCA	(PERM_CREATE|PERM_APPEND|PERM_READ|PERM_WRITE)

static patlist_t permtable[] = {
	{ "read",	PERM_READ                 , 0         },
	{ "write",	PERM_WRITE                , 0         },
	{ "append",	PERM_APPEND               , 0         },
	{ "create",	PERM_CREATE               , 0         },
	{ "permit",	PERM_READ   | PERM_WRITE  , PERM_CA   },
	{ "readonly",	PERM_READ                 , PERM_WCA  },
	{ "writeonly",	PERM_WRITE                , PERM_RCA  },
	{ "readwrite",	PERM_READ   | PERM_WRITE  , PERM_CA   },
	{ "deny",	PERM_DENY                 , PERM_RWCA },
	{ "discard",	PERM_DENY   | PERM_DISCARD, PERM_RWCA },
	{ NULL,		0                         , 0         } };


enum openModes { READ, WRITE, CREATE };

#define	DT_NONE		0
#define	DT_ANY		0
#define	DT_READ		1
#define DT_WRITE	2
#define DT_APPEND	3
#define DT_CREATE	4

/*
	translate system errno codes to apropriate tftp error codes
*/

int errno2p (int error)
{
#define PE_ACCESS       2               /* access violation */

#
    switch (error) {

        case EPERM:
        case EACCES:
        case EISDIR:
        case EROFS:
            return PX_ACCESS;

        case EEXIST:
            return PE_EXISTS;

        case ENOENT:
        case ENOTDIR:
        case ENODEV:
            return PX_NOTFOUND;
 
        case EFBIG:

       case ENOSPC:
           return PE_NOSPACE;
    }
    return PX_UNDEF;
}


/*
	parse access test command from '-a' command line option.
*/			

static int accTestParse (char *cmd, tftpR_t *rp)
{
    char *tok, *ct1, *ct2, *val=NULL;

    memset (rp, 0, sizeof(tftpR_t));

    for (tok=xstrtok(cmd,",",&ct1); tok!=NULL; tok=xstrtok(NULL,",",&ct1)){
        tok = xstrtok(tok, "=", &ct2);
        if ((tok==NULL) || (*(tok+1) != '\0')) {
            LOG_MSG1 (LOG_ERR, _("Illegal subcommand '%s'"), *tok);
            return 1;
        }
        switch (*tok) {
            case 'f':
            case 'F':
            case 's':
            case 'S':
            case 'c':
            case 'C':
                val = xstrtok(NULL,"=", &ct2);
                if (val==NULL) {
                    LOG_MSG1 (LOG_ERR, _("missing subcommand value '%s'"), *tok);
                    return 1;
                }
                break;
        }
        switch (*tok) {
            case 'w':
            case 'W':
                rp->reqType = RT_WRQ;
                break;
            case 'r':
            case 'R':
                rp->reqType = RT_RRQ;
                break;
            case 's':
            case 'S':
                if (sscanf(val, "%u", &rp->option.totalsize) != 1) {
                    LOG_MSG1 (LOG_ERR, _("missing subcommand value '%s'"), *tok);
                    return 1;
                }
                rp->option.types = ROT_TSZ;
                break;
            case 'f':
            case 'F':
                rp->reqFile = val;
                break;
            case 'c':
            case 'C':
                rp->reqPeer.sin_addr.s_addr = xinet_addr(val, RES_CONFIG);
                break;
        }
    }
    return 0;
}

/*
	check writability of file
*/

static int writeCheck (int ds, off_t offset)
{
    static char dummy=0;
    int nds;

    return 0;

    nds = dup(ds);
    if (lseek(nds, offset-1, SEEK_END)>=0)	/* position after EOF */
        if (write(nds, &dummy, 1) == 1) {	/* write one byte */
            close (nds);
            return 0;
        }

    close (nds);
    LOG_MSG1 (LOG_ERR, "%s %m", __FUNCTION__);
    return 1;
}

/*
	convert a textual permission list into a bitmask
*/

int evalPermission (char *pnames)
{
    int perm;

    perm = patstr(permtable, pnames, ",|:", PAT_EXACT|PAT_MULTI|PAT_NOCASE);
    return ((perm < 0) ? perm : (perm & PERM_EMASK));
}

/*
	Convert a permission mask into a comma separated text string
*/

char *perm2names (int permissionMask)
{
    if (!permissionMask) 
        return Strdup("none");

    return (bmap2pat (permissionMask, permtable, ","));
}

/*
	convert a direction type bitmask into a string
*/

static char *dtype2text (int dtype)
{
    switch (dtype) {
        case RT_RRQ:
            return "(output)";
        case RT_WRQ:
            return "(input)";
    }
    return "";
}


/*
	Rules:
	1.1)	In absence of any access list, grant read/write
	1.2)	find the client in the main access list.
	1.3)	if no matching entry exists at all, deny access anyway.

	2.1)	scan the file map list for a match of the requested file.
	2.2)	if found, replace the requested file name with the translated
		file name and proceed.
	
	3.1)	scan all restrict lists that match the number of the
		currently valid access list for a match of the requested file.
	3.2)	if no restrict list for that access list exists,
		grant access according to the access list entry.
	3.2)	if found a match, grant access according to the access 
		list entry too.
	3.3)	if no match found, proceed to the next access list entry.
*/

static int checkACL (tftpR_t *request)
{
    dla_t *ap;
    dlr_t *rp = NULL;
    dlf_t *fp;
    dlr_t *mp;
    dlt_t *tp;
    int pmask, apvalid=0;
    in_addr_t peer, m1st=0;

    request->aclNo = 0;

    peer  = (in_addr_t)request->reqPeer.sin_addr.s_addr;
    pmask = (request->reqType == RT_RRQ) ? PERM_READ :
                                           PERM_WRITE|PERM_CREATE|PERM_APPEND;
/*
	No access lists specified, fall back to traditional tftpd
*/
    if (isEmptyChain((dll_t *) accessListH)) {
        LOG_MSG2(LOG_WARNING, _("ACL empty: read,write permit"));
        return (PERM_READ | PERM_WRITE);
    }

/*
	get the most specific accesslist entry. 
	as this is not necessarily the only one, that's why we enter a loop !

	got to save the the netmask of each entry found to assure,
	so we won't deal with less specific entries later on.
*/
    for (ap=NULL; (ap=findACL(accessListH, ap, 0, peer, FIND_MAX)) != NULL; ) { 

        char *permbuf;

        permbuf = perm2names(ap->acl.perm);
        LOG_MSG4(LOG_INFO, _("ACL #%d: %s"), ap->mark, permbuf);
        if (permbuf != NULL)
            Free(permbuf);

        request->aclNo = ap->mark;
/*
	first round: if we get an explicit denial in the first round
*/
        if (!m1st) {
            if ((ap->acl.perm & PERM_DENY))
                return ap->acl.perm;
        }
/*
	second round or later: less specific means- we're done
*/
	else {
            if (m1st > ap->acl.mask)
                break;
        }

        m1st = ap->acl.mask;
/*
	scan the restrict lists - this is where we might need annother round:
	if we don't find a match for the requested file name,
	why not test other access lists for a match ?
*/
        if ((rp = getReByMark(restrictListH, NULL, ap->mark)) == NULL) {
            LOG_MSG4(LOG_INFO, _("ACL #%d: no restrictions"), ap->mark);
            apvalid = 1;
            break;
        }
        if ((rp = findRePattern(restrictListH, NULL, ap->mark,
                                request->reqFile, request->reqType)) != NULL) {
            LOG_MSG4(LOG_INFO, _("ACL #%d: permit '%s'=>'%s'"),
                                ap->mark, request->reqFile, rp->raw);
            apvalid = 1;
            break;
        }
    }

    if (!apvalid) {
        LOG_MSG4(LOG_INFO, _("No matching access/restrict list found"));
        return PERM_NONE;
    }

/*
	scan the file map lists to see if we need a name translation
*/
    if ((mp = findRePattern(mapListH, NULL, ap->mark, 
                            request->reqFile, request->reqType)) != NULL) {
        request->realFile = xlateFileMap(request->reqFile, mp->map,
                                         &request->reqPeer, NULL,
                                         request->reqType);
        if (request->realFile != NULL) {
            LOG_MSG4(LOG_INFO, _("Map #%d %s from '%s' to '%s'"),
                                mp->mark, dtype2text(mp->dtype),
                                request->reqFile, request->realFile);
        }
        else {
            return PERM_NONE;
        }
    }
    else {
        request->realFile = Strdup(request->reqFile);
        if (request->realFile == NULL) {
            errNoDynmem;
            return PERM_NONE;
        }
    }

/*
	now, that the access is granted anyway,
	check here if we are intended to use i/o filtering
*/
    if ((rp != NULL) && (needFilter = rp->filter)) {
        if ((fp = getFilterByMark(filterListH, NULL, needFilter)) != NULL) {
            if (request->reqType == RT_RRQ) {
                if (fp->rxfilter != NULL) {
                    request->filter = fp->rxfilter;
                }
                else {
                    needFilter = 0;
                }
            }
            else {
                if (fp->txfilter != NULL) {
                    request->filter = fp->txfilter;
                }
                else {
                    needFilter = 0;
                }
            }
        } 
        else {
            needFilter = 0;
        }
        if (needFilter) {
            char *tempFilter;
            tempFilter = xlateFileMap(request->reqFile, request->filter,
                                      &request->reqPeer, request->realFile,
                                      request->reqType);
            if (tempFilter != NULL) {
                LOG_MSG4(LOG_INFO, _("Applying i/o filter #%d '%s'"),
                                    fp->mark, tempFilter);
                Free(tempFilter);
            }
        }
    }

/*
	finally look for a connection specific timer chain entry
*/
 
    if ((tp=getTimerByMark(timerListH, NULL, ap->mark)) != NULL) {
        request->times = tp->times;
        LOG_MSG4(LOG_INFO, _("ACL #%d: Timing: %s"), ap->mark,
                            timer2asc(&request->times, T2A_XFER|T2A_RET));
    }

    if (rp != NULL) {
         LOG_MSG4(LOG_INFO, _("ACL #%d: Maximum Transfer size: %ld"), 
                  ap->mark, rp->maxsize); 
         request->maxsize = rp->maxsize;
     }
     else
         request->maxsize = 0;

    return (ap->acl.perm & pmask);
}

static void denyLog (int dt, char *text2)
{
    char *text1 = "", *sep = ": ";

    if (!safeLog) {
      switch (dt) {
        case DT_READ:
            text1 = _("Read ");
            break;
        case DT_WRITE:
            text1 = _("Write ");
            break;
        case DT_CREATE:
            text1 = _("Create ");
            break;
        default:
            text1 = "";
            break;
      }
      if (text2 == NULL) {
          sep   = "";
          text2 = "";
      }
    }
    else {
        sep   = "";
        text2 = "";
    }

    LOG_MSG2 (LOG_WARNING, _("%saccess denied%s%s"), text1, sep, text2);
}

static char *copenMode(enum openModes om, int appflag)
{
    switch (om) {
        case READ:
            return _("READ");
        case WRITE:
            if (appflag)
                return _("APPEND");
            else
                return _("WRITE");
        case CREATE:
            return _("CREATE");
    }
    return "";
}

/*
	Validate file access on basis of request (transfer) type, 
	requested file and the peer's ip address.
	returns 0 and a FILE* to the opened requestd file
	on success or an TFTP error code else.
*/

int validateAccess (tftpR_t *request)
{
    struct stat stbuf;
    enum openModes openMode = READ;
    int perm, rtc, ios, ds, omask, cmode; 
    char *fname, *temp;
#ifdef FILE_CACHE
    dlc_t *entry;
#endif /* FILE_CACHE */

/*
	If this is an access test run, evaluate our faked request
*/

    if (accTest != NULL)
        if (accTestParse(accTest, request))
            return 1;

/*
	Evaluate the access lists
	fills in the 'real' (translated) file name too
*/

    perm = checkACL(request);

/*
	The requested operation must match the current access lists
*/
    rtc = ((perm & PERM_DISCARD)) ? PX_DISCARD : PX_ACCESS;

    if ((request->reqType == RT_RRQ) && !(perm & PERM_READ)) {
        denyLog (DT_READ, _("by ACL"));
        return rtc;
    }
    else if ((request->reqType == RT_WRQ) && !(perm & PERM_WRCRAP)) {
        denyLog (DT_ANY, _("by ACL"));
        return rtc;
    }
    else if (!safeLog) {
        LOG_MSG4 (LOG_INFO, _("ACL access possible to %s"), request->realFile);
    }

/*
	Check for a xfer-size limit
*/
    if (request->maxsize) {
        if (request->option.types & ROT_TSZ) {
            if (request->option.totalsize > request->maxsize) {
                LOG_MSG2 (LOG_WARNING, 
                          _("ACL Transfer size limit (%ld) exceeded"), 
                          request->maxsize);
                return PX_NOSPACE;
            }
        }
        else {
            request->option.totalsize = request->maxsize;
        }
        LOG_MSG5 (LOG_DEBUG, "Transfer size limit: %ld",
                  request->option.totalsize);
    }
    else {
        request->option.totalsize = 0;
    }
              
/*
	now evaluate the requested file name into a resolved path
	removing all self relative constructs and symlinks

	read request: 	as the requested file must exist,
			realPath must return without error.

	write request:	the file may or may not exist yet
			we allow realPath to return with ENOENT
			which means the file was not found but the
			directory path to the file is valid & accessible.
			the parent directory must be writable in order to
			create a new file.
*/
    ios = realPath (defaultDir, request->realFile, &temp, 0);

    if (ios == ENOMEM)
        return PX_INTERNAL;

    if (request->realFile != NULL)
        Free(request->realFile);
    request->realFile = temp;

    if (request->reqType == RT_RRQ) {
        if (ios) {
            denyLog (DT_READ, strerror(ios));
            return errno2p(ios);

        }
        if (stat(request->realFile, &stbuf) || !S_ISREG(stbuf.st_mode)) {
            denyLog (DT_NONE, _("not a regular file"));
            return PX_ACCESS;
        }
        openMode = READ;
    }
    else { 			/* request->reqType == RT_WRQ */
        switch (ios) {
            case 0:
                if (fileAccess(request->realFile, ACC_WRITE)) {
                    denyLog (DT_WRITE, _("no write permission"));
                    return PX_ACCESS;
                }
                if (!(perm & PERM_WRITE)) {
                    denyLog (DT_WRITE, _("file exists"));
                    return PX_ACCESS;
                }
                openMode = WRITE;
                break;
            case ENOENT:
                if (fileAccess(request->realFile, ACC_CREATE)) {
                    denyLog (DT_CREATE, _("no write access to directory"));
                    return PX_ACCESS;
                }
                if (!(perm & PERM_CREATE)) {
                    denyLog (DT_CREATE, NULL);
                    return PX_ACCESS;
                }
                openMode = CREATE;
                break;
            default:
                denyLog (DT_ANY, strerror(ios));
                return errno2p(ios);
        }
    }

/*
	if there's a path restriction (via: DefaultDirectory !<pathname>)
	deny anything outside this tree
*/
    if (keepBelow) {
        if (isInTree(defaultDir, request->realFile)) {
            denyLog (DT_NONE, _("outside default directory tree"));
            return PX_ACCESS;
        }
    }
            
/*
	Deny access to our own config file anyway
*/
    if ((ios = realPath(NULL, configFile, &temp, 0))) {
        LOG_MSG1 (LOG_ERR, "%s", temp);
        LOG_MSG1 (LOG_ERR, _("Cannot get real path of config file: %s"), 
                  strerror(ios));
        if (ios != ENOMEM)
            Free(temp);
        return PX_INTERNAL;
    }
    if (!strcmp(request->realFile, temp)) {
        denyLog (DT_NONE, _("config file"));
        if (ios != ENOMEM)
            Free(temp);
        return PX_ACCESS;
    }
    if (ios != ENOMEM)
        Free(temp);
/*
	at this point access to the requested file is granted
	try to open / create the file
*/

    if (accTest != NULL)		/* access test ends here	*/
        return 0;
                
    switch (openMode) {

        case READ:

#ifdef FILE_CACHE
/*
	file caching is only valid without active filters. this is because
	we cannot determine at preload time, what client (resp. filter) 
	requests the file

	so, check if we got a preloaded file.
	if so, we don't need to run through the open() section.
*/
            if ((entry = findCacheByMark(cacheListH, request->realFile,
                                         request->reqMode)) != NULL) {
                if (needFilter) {
                    LOG_MSG2 (LOG_WARNING, 
                              _("Cannot use cached file: filter needed"));
                }
                else if (entry->flags & CAC_VALID) {
                    LOG_MSG4 (LOG_INFO, _("Cache hit"));
                    request->option.totalsize = entry->status.st_size;
                    request->fcache = (char *)entry;
                    entry->used = time(NULL);
                    break;
                }
                else {
                    LOG_MSG5 (LOG_DEBUG, "File not preloaded");
                }
            }
            else {
                LOG_MSG5 (LOG_DEBUG, "File not cached");
            }

#endif /* FILE_CACHE */

            ds = open (request->realFile, O_RDONLY);
 
            if (ds >= 0) {
                struct stat st;

                if (fstat(ds, &st)) {
                    LOG_MSG1 (LOG_ERR, _("Cannot stat file '%s'"),
                              request->realFile);
                    close (ds);
                    return PX_ACCESS;
                }
                else {
                    request->option.totalsize = st.st_size;
                }

                if (needFilter) {
                    if (forkFilter(request, &ds)) {
                        close (ds);
                        return PX_ACCESS;
                    }
                }
                request->reqFd = ds;
            }
            else
                request->reqFd = -1;
            break;

        case WRITE:
            if (perm & PERM_APPEND) {
                if (needFilter) {
                    LOG_MSG1 (LOG_ERR, _("Cannot APPEND using a filter"));
                    return PX_ACCESS;
                }

                ds = open (request->realFile, O_WRONLY | O_APPEND);

                if ((ds >= 0) && xferOptions && (request->option.types & ROT_TSZ)) {
                    if (writeCheck(request->reqFd, request->option.totalsize)) {
                        close (ds);
                        return PX_NOSPACE;
                    }
                }

                if (ds >= 0)
                    request->reqFd = ds;
                else
                    request->reqFd = -1;
            }
            else {
                ds = open (request->realFile, O_WRONLY | O_TRUNC);

                if ((ds >= 0) && xferOptions && (request->option.types & ROT_TSZ)) {
                    if (writeCheck(request->reqFd, request->option.totalsize)) {
                        close (ds);
                        return PX_NOSPACE;
                    }
                }

                if (ds >= 0) {
                    if (needFilter) {
                        if (forkFilter (request, &ds)) {
                            close (ds);
                            return PX_ACCESS;
                        }
                    }
                    request->reqFd = ds;
                }
                else
                    request->reqFd = -1;
            }
            break;
  
        case CREATE:

            cmode = S_IRUSR | S_IRGRP;
            if (perm & (PERM_WRITE|PERM_APPEND)) cmode |= (S_IWUSR | S_IWGRP);

            omask = umask (0);

            LOG_MSG5 (LOG_DEBUG, "Create mode: %o", cmode);
            
            ds = open (request->realFile, 
                       O_WRONLY | O_CREAT | O_TRUNC | O_EXCL, createMode);

            if ((ds >= 0) && xferOptions && (request->option.types & ROT_TSZ)) {
                if (writeCheck(request->reqFd, request->option.totalsize)) {
                    close (ds);
                    return PX_NOSPACE;
                }
            }

            if (ds >= 0) {
                (void) umask (omask);

                if (needFilter) {
                    if (forkFilter (request, &ds)) {
                        close (ds);
                        return PX_ACCESS;
                    }
                }
                request->reqFd = ds;
            }
            else
                request->reqFd = -1;
            break;
    }

#ifdef FILE_CACHE

    if (request->fcache != NULL)
        return (0);

#endif /* FILE_CACHE */

    if (request->reqFd < 0) {
        LOG_MSG2 (LOG_WARNING, _("fd/open (%s) fails for %s: %m"), 
                  copenMode(openMode, perm & PERM_APPEND),
                  request->realFile, strerror(errno));
        return PX_ACCESS;
    }
    fname = baseName(request->filter);
    if (needFilter) {
        LOG_MSG4 (LOG_INFO, _("File '%s' opened for %s via filter '%s'"), 
                  request->realFile,copenMode(openMode,perm & PERM_APPEND),
                  fname);
    }
    else {
        LOG_MSG4 (LOG_INFO, _("File '%s' opened for %s"), 
                  request->realFile,copenMode(openMode,perm & PERM_APPEND));
    }
    return (0);
}

/*
	check thread restrictions

	returns the number of allowed threads remaining in this group,
	-1 if no restrictions apply.
*/

int checkLimits(tftpR_t *req)
{
    dla_t *da = NULL;
    int used;
    in_addr_t peer;
/*
	scan the thread restriction list for a matching entry.
	if nothing matching found, we assume -no- restriction
*/
    peer = (in_addr_t)req->reqPeer.sin_addr.s_addr;

    if ((da = findACL(threadListH, NULL, 0, peer, FIND_MAX)) == NULL) {
        LOG_MSG5 (LOG_DEBUG, "No thread limit for %s", 
                  xinet_ntoa(peer, doResolv & RES_RUNTIME));
        return -1;
    }
/*
	now we know, how many threads of this address mask are allowed
	we're going to scan the process list for entries of this type
*/
     SIGBLOCK(SIGCHLD, (used = getPidCountByMark(procListH, da->mark)));

    if (used < da->acl.perm) {
        struct in_addr sia;
        sia.s_addr = da->acl.addr & da->acl.mask;
        LOG_MSG5 (LOG_DEBUG, "%d of %d slots left for %s",
                 (da->acl.perm-used), da->acl.perm, inet_ntoa(sia));
    }
    else {
        LOG_MSG2 (LOG_WARNING, _("Limit %d reached for %s"),
                  da->acl.perm, xinet_ntoa(peer, doResolv & RES_RUNTIME));
        return 0;
    }

    return da->mark;
}

